import base64
from urllib.parse import unquote


def process_xop(document, message_pack):
    """Iterate through the tree and replace the xop:include elements."""

    xop_nodes = document.xpath(
        "//xop:Include", namespaces={"xop": "http://www.w3.org/2004/08/xop/include"}
    )
    num_replaced = 0

    for xop_node in xop_nodes:
        href = xop_node.get("href")
        if href.startswith("cid:"):
            # URL can be encoded. RFC2392
            href = "<%s>" % unquote(href[4:])

        value = message_pack.get_by_content_id(href)
        if not value:
            raise ValueError("No part found for: %r" % xop_node.get("href"))
        num_replaced += 1

        xop_parent = xop_node.getparent()
        xop_parent.remove(xop_node)
        xop_parent.text = base64.b64encode(value.content)

    return num_replaced > 0



import logging

from lxml import etree

from zeep import exceptions, ns
from zeep.utils import qname_attr
from zeep.xsd.const import NotSet, xsi_ns
from zeep.xsd.elements.base import Base
from zeep.xsd.utils import max_occurs_iter
from zeep.xsd.valueobjects import AnyObject

logger = logging.getLogger(__name__)


__all__ = ["Any", "AnyAttribute"]


class Any(Base):
    name = None

    def __init__(
        self, max_occurs=1, min_occurs=1, process_contents="strict", restrict=None
    ):
        """

        :param process_contents: Specifies how the XML processor should handle
                                 validation against the elements specified by
                                 this any element
        :type process_contents: str (strict, lax, skip)

        """
        super().__init__()
        self.max_occurs = max_occurs
        self.min_occurs = min_occurs
        self.restrict = restrict
        self.process_contents = process_contents

        # cyclic import
        from zeep.xsd import AnyType

        self.type = AnyType()

    def __call__(self, any_object):
        return any_object

    def __repr__(self):
        return "<%s(name=%r)>" % (self.__class__.__name__, self.name)

    def accept(self, value):
        return True

    def parse(self, xmlelement, schema, context=None):
        if self.process_contents == "skip":
            return xmlelement

        # If a schema was passed inline then check for a matching one
        qname = etree.QName(xmlelement.tag)
        if context and context.schemas:
            for context_schema in context.schemas:
                if context_schema.documents.has_schema_document_for_ns(qname.namespace):
                    schema = context_schema
                    break
            else:
                # Try to parse the any result by iterating all the schemas
                for context_schema in context.schemas:
                    try:
                        data = context_schema.deserialize(list(xmlelement)[0])
                        return data
                    except LookupError:
                        continue

        # Lookup type via xsi:type attribute
        xsd_type = qname_attr(xmlelement, xsi_ns("type"))
        if xsd_type is not None:
            xsd_type = schema.get_type(xsd_type)
            return xsd_type.parse_xmlelement(xmlelement, schema, context=context)

        # Check if a restrict is used
        if self.restrict:
            return self.restrict.parse_xmlelement(xmlelement, schema, context=context)

        try:
            element = schema.get_element(xmlelement.tag)
            return element.parse(xmlelement, schema, context=context)
        except (exceptions.NamespaceError, exceptions.LookupError):
            return xmlelement

    def parse_kwargs(self, kwargs, name, available_kwargs):
        if name in available_kwargs:
            available_kwargs.remove(name)
            value = kwargs[name]
            return {name: value}
        return {}

    def parse_xmlelements(self, xmlelements, schema, name=None, context=None):
        """Consume matching xmlelements and call parse() on each of them

        :param xmlelements: Dequeue of XML element objects
        :type xmlelements: collections.deque of lxml.etree._Element
        :param schema: The parent XML schema
        :type schema: zeep.xsd.Schema
        :param name: The name of the parent element
        :type name: str
        :param context: Optional parsing context (for inline schemas)
        :type context: zeep.xsd.context.XmlParserContext
        :return: dict or None

        """
        result = []

        for _unused in max_occurs_iter(self.max_occurs):
            if xmlelements:
                xmlelement = xmlelements.popleft()
                item = self.parse(xmlelement, schema, context=context)
                if item is not None:
                    result.append(item)
            else:
                break

        if not self.accepts_multiple:
            result = result[0] if result else None
        return result

    def render(self, parent, value, render_path=None):
        assert parent is not None
        self.validate(value, render_path)

        if self.accepts_multiple and isinstance(value, list):
            from zeep.xsd import AnySimpleType

            if isinstance(self.restrict, AnySimpleType):
                for val in value:
                    node = etree.SubElement(parent, "item")
                    node.set(xsi_ns("type"), self.restrict.qname)
                    self._render_value_item(node, val, render_path)
            elif self.restrict:
                for val in value:
                    node = etree.SubElement(parent, self.restrict.name)
                    # node.set(xsi_ns('type'), self.restrict.qname)
                    self._render_value_item(node, val, render_path)
            else:
                for val in value:
                    self._render_value_item(parent, val, render_path)
        else:
            self._render_value_item(parent, value, render_path)

    def _render_value_item(self, parent, value, render_path):
        if value in (None, NotSet):  # can be an lxml element
            return

        elif isinstance(value, etree._Element):
            parent.append(value)

        elif self.restrict:
            if isinstance(value, list):
                for val in value:
                    self.restrict.render(parent, val, None, render_path=render_path)
            else:
                self.restrict.render(parent, value, None, render_path=render_path)
        else:
            if isinstance(value.value, list):
                for val in value.value:
                    value.xsd_elm.render(parent, val, render_path=render_path)
            else:
                value.xsd_elm.render(parent, value.value, render_path=render_path)

    def validate(self, value, render_path):
        if self.accepts_multiple and isinstance(value, list):

            # Validate bounds
            if len(value) < self.min_occurs:
                raise exceptions.ValidationError(
                    "Expected at least %d items (minOccurs check)" % self.min_occurs
                )
            if (
                self.max_occurs != "unbounded"
                and isinstance(self.max_occurs, int)
                and len(value) > self.max_occurs
            ):
                raise exceptions.ValidationError(
                    "Expected at most %d items (maxOccurs check)" % self.min_occurs
                )

            for val in value:
                self._validate_item(val, render_path)
        else:
            if not self.is_optional and value in (None, NotSet):
                raise exceptions.ValidationError("Missing element for Any")

            self._validate_item(value, render_path)

    def _validate_item(self, value, render_path):
        if value is None:  # can be an lxml element
            return

        # Check if we received a proper value object. If we receive the wrong
        # type then return a nice error message
        if self.restrict:
            expected_types = [etree._Element, dict] + self.restrict.accepted_types
        else:
            expected_types = [etree._Element, dict, AnyObject]

        if value in (None, NotSet):
            if not self.is_optional:
                raise exceptions.ValidationError(
                    "Missing element %s" % (self.name), path=render_path
                )

        elif not isinstance(value, tuple(expected_types)):
            type_names = ["%s.%s" % (t.__module__, t.__name__) for t in expected_types]
            err_message = "Any element received object of type %r, expected %s" % (
                type(value).__name__,
                " or ".join(type_names),
            )

            raise TypeError(
                "\n".join(
                    (
                        err_message,
                        "See http://docs.python-zeep.org/en/master/datastructures.html"
                        "#any-objects for more information",
                    )
                )
            )

    def resolve(self):
        return self

    def signature(self, schema=None, standalone=True):
        if self.restrict:
            base = self.restrict.name
        else:
            base = "ANY"

        if self.accepts_multiple:
            return "%s[]" % base
        return base


class AnyAttribute(Base):
    # FIXME: should not inherit from Base
    name = None
    _ignore_attributes = [etree.QName(ns.XSI, "type")]

    def __init__(self, process_contents="strict"):
        self.qname = None
        self.process_contents = process_contents

    def parse(self, attributes, context=None):
        result = {}
        for key, value in attributes.items():
            if key not in self._ignore_attributes:
                result[key] = value
        return result

    def resolve(self):
        return self

    def render(self, parent, value, render_path=None):
        if value in (None, NotSet):
            return

        for name, val in value.items():
            parent.set(name, val)

    def signature(self, schema=None, standalone=True):
        return "{}"



"""
    zeep.wsdl.definitions
    ~~~~~~~~~~~~~~~~~~~~~

    A WSDL document exists out of a number of definitions. There are 6 major
    definitions, these are:

     - types
     - message
     - portType
     - binding
     - port
     - service

    This module defines the definitions which occur within a WSDL document,

"""
import typing
import warnings
from collections import OrderedDict, namedtuple

from lxml import etree

from zeep.exceptions import IncompleteOperation

if typing.TYPE_CHECKING:
    from zeep.wsdl.wsdl import Definition

MessagePart = namedtuple("MessagePart", ["element", "type"])


class AbstractMessage:
    """Messages consist of one or more logical parts.

    Each part is associated with a type from some type system using a
    message-typing attribute. The set of message-typing attributes is
    extensible. WSDL defines several such message-typing attributes for use
    with XSD:

        - element: Refers to an XSD element using a QName.
        - type: Refers to an XSD simpleType or complexType using a QName.

    """

    def __init__(self, name):
        self.name = name
        self.parts = OrderedDict()

    def __repr__(self):
        return "<%s(name=%r)>" % (self.__class__.__name__, self.name.text)

    def resolve(self, definitions):
        pass

    def add_part(self, name, element):
        self.parts[name] = element


class AbstractOperation:
    """Abstract operations are defined in the wsdl's portType elements."""

    def __init__(
        self,
        name,
        input_message=None,
        output_message=None,
        fault_messages=None,
        parameter_order=None,
        wsa_action=None,
    ):
        """Initialize the abstract operation.

        :param name: The name of the operation
        :type name: str
        :param input_message: Message to generate the request XML
        :type input_message: AbstractMessage
        :param output_message: Message to process the response XML
        :type output_message: AbstractMessage
        :param fault_messages: Dict of messages to handle faults
        :type fault_messages: dict of str: AbstractMessage

        """
        self.name = name
        self.input_message = input_message
        self.output_message = output_message
        self.fault_messages = fault_messages
        self.parameter_order = parameter_order
        self.wsa_action = wsa_action


class PortType:
    def __init__(
        self, name: etree.QName, operations: typing.Dict[str, AbstractOperation]
    ):
        self.name = name
        self.operations = operations

    def __repr__(self):
        return "<%s(name=%r)>" % (self.__class__.__name__, self.name.text)

    def resolve(self, definitions):
        pass


class Binding:
    """Base class for the various bindings (SoapBinding / HttpBinding)

    .. raw:: ascii

        Binding
           |
           +-> Operation
                   |
                   +-> ConcreteMessage
                             |
                             +-> AbstractMessage

    """

    def __init__(self, wsdl, name, port_name):
        """Binding

        :param wsdl:
        :type wsdl:
        :param name:
        :type name: string
        :param port_name:
        :type port_name: string

        """
        self.name = name
        self.port_name = port_name
        self.port_type = None
        self.wsdl = wsdl
        self._operations = {}

    def resolve(self, definitions: "Definition") -> None:
        self.port_type = definitions.get("port_types", self.port_name.text)

        for name, operation in list(self._operations.items()):
            try:
                operation.resolve(definitions)
            except IncompleteOperation as exc:
                warnings.warn(str(exc))
                del self._operations[name]

    def _operation_add(self, operation):
        # XXX: operation name is not unique
        self._operations[operation.name] = operation

    def __str__(self):
        return "%s: %s" % (self.__class__.__name__, self.name.text)

    def __repr__(self):
        return "<%s(name=%r, port_type=%r)>" % (
            self.__class__.__name__,
            self.name.text,
            self.port_type,
        )

    def all(self):
        return self._operations

    def get(self, key):
        try:
            return self._operations[key]
        except KeyError:
            raise ValueError("No such operation %r on %s" % (key, self.name))

    @classmethod
    def match(cls, node):
        raise NotImplementedError()

    @classmethod
    def parse(cls, definitions, xmlelement):
        raise NotImplementedError()


class Operation:
    """Concrete operation

    Contains references to the concrete messages

    """

    def __init__(self, name, binding):
        self.name = name
        self.binding = binding
        self.abstract = None
        self.style = None
        self.input = None
        self.output = None
        self.faults = {}

    def resolve(self, definitions):
        try:
            self.abstract = self.binding.port_type.operations[self.name]
        except KeyError:
            raise IncompleteOperation(
                "The wsdl:operation %r was not found in the wsdl:portType %r"
                % (self.name, self.binding.port_type.name.text)
            )

    def __repr__(self):
        return "<%s(name=%r, style=%r)>" % (
            self.__class__.__name__,
            self.name,
            self.style,
        )

    def __str__(self):
        if not self.input:
            return u"%s(missing input message)" % (self.name)

        retval = u"%s(%s)" % (self.name, self.input.signature())
        if self.output:
            retval += u" -> %s" % (self.output.signature(as_output=True))
        return retval

    def create(self, *args, **kwargs):
        assert self.input is not None
        return self.input.serialize(*args, **kwargs)

    def process_reply(self, envelope):
        raise NotImplementedError()

    @classmethod
    def parse(cls, wsdl, xmlelement, binding):
        """

        Definition::

            <wsdl:operation name="nmtoken"> *
               <-- extensibility element (2) --> *
               <wsdl:input name="nmtoken"? > ?
                   <-- extensibility element (3) -->
               </wsdl:input>
               <wsdl:output name="nmtoken"? > ?
                   <-- extensibility element (4) --> *
               </wsdl:output>
               <wsdl:fault name="nmtoken"> *
                   <-- extensibility element (5) --> *
               </wsdl:fault>
            </wsdl:operation>

        """
        raise NotImplementedError()


class Port:
    """Specifies an address for a binding, thus defining a single communication
    endpoint.

    """

    if typing.TYPE_CHECKING:
        _resolve_context = None  # type: typing.Optional[typing.Dict[str, typing.Any]]

    def __init__(self, name, binding_name, xmlelement):
        self.name = name
        self._resolve_context = {"binding_name": binding_name, "xmlelement": xmlelement}

        # Set during resolve()
        self.binding = None
        self.binding_options = {}

    def __repr__(self):
        return "<%s(name=%r, binding=%r, %r)>" % (
            self.__class__.__name__,
            self.name,
            self.binding,
            self.binding_options,
        )

    def __str__(self):
        return u"Port: %s (%s)" % (self.name, self.binding)

    def resolve(self, definitions):
        if self._resolve_context is None:
            return

        try:
            self.binding = definitions.get(
                "bindings", self._resolve_context["binding_name"].text
            )
        except IndexError:
            return False

        if definitions.location and self.binding.wsdl.settings.force_https:
            force_https = definitions.location.startswith("https")
        else:
            force_https = False

        self.binding_options = self.binding.process_service_port(
            self._resolve_context["xmlelement"], force_https
        )
        self._resolve_context = None
        return True


class Service:
    """Used to aggregate a set of related ports."""

    def __init__(self, name):
        self.ports = OrderedDict()
        self.name = name
        self._is_resolved = False

    def __str__(self):
        return u"Service: %s" % self.name

    def __repr__(self):
        return "<%s(name=%r, ports=%r)>" % (
            self.__class__.__name__,
            self.name,
            self.ports,
        )

    def resolve(self, definitions):
        if self._is_resolved:
            return

        unresolved = []
        for name, port in self.ports.items():
            is_resolved = port.resolve(definitions)
            if not is_resolved:
                unresolved.append(name)

        # Remove unresolved bindings (http etc)
        for name in unresolved:
            del self.ports[name]

        self._is_resolved = True

    def add_port(self, port: Port) -> None:
        self.ports[port.name] = port



from collections import OrderedDict
from io import StringIO


class PrettyPrinter:
    """Cleaner pprint output.

    Heavily inspired by the Python pprint module, but more basic for now.

    """

    def pformat(self, obj):
        stream = StringIO()
        self._format(obj, stream)
        return stream.getvalue()

    def _format(self, obj, stream, indent=4, level=1):
        _repr = getattr(type(obj), "__repr__", None)
        write = stream.write

        if (isinstance(obj, dict) and _repr is dict.__repr__) or (
            isinstance(obj, OrderedDict) and _repr == OrderedDict.__repr__
        ):
            write("{\n")
            num = len(obj)

            if num > 0:
                for i, (key, value) in enumerate(obj.items()):
                    write(" " * (indent * level))
                    write("'%s'" % key)
                    write(": ")
                    self._format(value, stream, level=level + 1)
                    if i < num - 1:
                        write(",")
                    write("\n")

                write(" " * (indent * (level - 1)))
            write("}")

        elif isinstance(obj, list) and _repr is list.__repr__:
            write("[")
            num = len(obj)

            if num > 0:
                write("\n")
                for i, value in enumerate(obj):
                    write(" " * (indent * level))
                    self._format(value, stream, level=level + 1)
                    if i < num - 1:
                        write(",")
                    write("\n")
                write(" " * (indent * (level - 1)))
            write("]")
        else:
            value = repr(obj)
            if "\n" in value:
                lines = value.split("\n")
                num = len(lines)
                for i, line in enumerate(lines):
                    if i > 0:
                        write(" " * (indent * (level - 1)))
                    write(line)
                    if i < num - 1:
                        write("\n")
            else:
                write(value)
